"use strict";
var __read = (this && this.__read) || function (o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.parseSpdxExpressionWithDetails = exports.parse = void 0;
var strict_parser_1 = require("./strict_parser");
var liberal_parser_1 = require("./liberal_parser");
var licenses_1 = require("../licenses");
/**
 * Clean up the given SPDX expression before processing it with the grammar.
 *
 * Currently, the clean-up is limited to:
 * 1. trimming the surrounding whitespace, and
 * 2. compressing consecutive whitespace characters into a single space.
 *
 * @param input SPDX expression as a string
 * @param strictSyntax If {true}, don't clean up spec-violating details
 * @returns A slightly cleaned up SPDX expression as a string
 */
var prepareInput = function (input, strictSyntax) {
    var cleanedInput = input
        .trim() // remove surrounding whitespace
        .replace(/\s+/, ' '); // compress consecutive whitespaces to a single space
    if (!strictSyntax) {
        cleanedInput = cleanedInput
            .replace(/\s+and\s+/i, ' AND ') // fix lowercase keywords
            .replace(/\s+or\s+/i, ' OR ') // fix lowercase keywords
            .replace(/\s+with\s+/i, ' WITH '); // fix lowercase keywords
    }
    return cleanedInput;
};
var compileFullSpdxParseResult = function (input, parserResult) {
    var _a;
    return { input: input, ast: (_a = parserResult.parse) === null || _a === void 0 ? void 0 : _a.ast, error: parserResult.error, expression: parserResult.expression };
};
var buildErrorMessage = function (input, strictSyntax) {
    return "".concat(strictSyntax ? 'Strict' : 'Liberal', " parsing for ").concat(JSON.stringify(input), " failed");
};
/**
 * Parse an SPDX expression into a structured object representation.
 *
 * @param input SPDX expression as a string to be parsed.
 * @returns {ParsedSpdxExpression} A structured object describing the given SPDX expression
 *          or throws an `Error` if parsing failed.
 */
function parse(input, strictSyntax) {
    if (strictSyntax === void 0) { strictSyntax = false; }
    var data = parseSpdxExpressionWithDetails(input, strictSyntax);
    if (data.error) {
        throw new Error([buildErrorMessage(input, strictSyntax), data.error].join(': '));
    }
    else if (data.expression === undefined) {
        /* istanbul ignore next */ throw new Error(buildErrorMessage(input, strictSyntax));
    }
    return data.expression;
}
exports.parse = parse;
/**
 * Parse an SPDX expression into a structured object representation along with additional
 * metadata such as the underlying AST tree used and any errors in case parsing failed.
 *
 * @param input SPDX expression as a string to be parsed.
 * @returns {FullSpdxParseResult}
 */
function parseSpdxExpressionWithDetails(input, strictSyntax) {
    if (strictSyntax === void 0) { strictSyntax = false; }
    // Apply general clean up on the raw input string
    var preparedInput = prepareInput(input, strictSyntax);
    // Always try to parse with the strict parser first in order to
    // minimize risk of unwanted "corrections"
    var notCorrected = (0, strict_parser_1.parse)(preparedInput);
    if (notCorrected.expression && !strictSyntax) {
        notCorrected.expression = correctDashSeparatedLicenseIds(notCorrected.expression);
    }
    var strictResult = compileFullSpdxParseResult(preparedInput, notCorrected);
    // If strict parsing failed, attempt to apply corrections if allowed
    if (strictResult.error && !strictSyntax) {
        var corrected = (0, liberal_parser_1.parse)(preparedInput);
        if (corrected.expression) {
            corrected.expression = correctDashSeparatedLicenseIds(corrected.expression);
        }
        var liberalResult = compileFullSpdxParseResult(preparedInput, corrected);
        // We'll use the results of liberal parsing only if it succeeds. In case
        // of failure, we'll return the errors from strict parser for maximum
        // awareness of deviations from the SPDX specification
        if (!liberalResult.error) {
            return liberalResult;
        }
    }
    // If liberal parsing was not allowed or it failed, too, let's return the
    // (possibly failed) strict parsing result
    return strictResult;
}
exports.parseSpdxExpressionWithDetails = parseSpdxExpressionWithDetails;
var correctDashSeparatedLicenseIds = function (expression) {
    var recurse = function (expression) {
        var fixLicenseRef = function (node) { return node; };
        var fixConjunctionInfo = function (node) {
            return { conjunction: node.conjunction, left: recurse(node.left), right: recurse(node.right) };
        };
        var fixLicenseInfo = function (node) {
            if (!node.exception && node.license.includes('-with-') && !(0, licenses_1.isKnownLicenseId)(node.license)) {
                // let's see if we can map the parts to actual known identifiers
                var _a = __read(node.license.split('-with-'), 2), splitLicense = _a[0], splitException = _a[1];
                var fixedLicense = (0, licenses_1.correctLicenseId)(splitLicense);
                var fixedException = (0, licenses_1.correctExceptionId)(splitException, fixedLicense);
                if ((0, licenses_1.isKnownLicenseId)(fixedLicense)) {
                    return { license: fixedLicense, exception: fixedException };
                }
            }
            return node;
        };
        if (expression.conjunction) {
            return fixConjunctionInfo(expression);
        }
        else if (expression.licenseRef) {
            return fixLicenseRef(expression);
        }
        else {
            return fixLicenseInfo(expression);
        }
    };
    return recurse(expression);
};
