"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.validate = void 0;
var assert_1 = require("assert");
var parser_1 = require("../parser");
var licenses_1 = require("../licenses");
var ALL_LICENSE_IDS = licenses_1.licenses.licenses.map(function (entry) { return entry.licenseId; });
var LOWERCASE_LICENSE_IDS = ALL_LICENSE_IDS.map(function (id) { return id.toLowerCase(); });
var LOWERCASE_EXCEPTION_IDS = licenses_1.exceptions.exceptions.map(function (entry) { return entry.licenseExceptionId.toLowerCase(); });
var isKnownLicenseIdentifier = function (id) { return LOWERCASE_LICENSE_IDS.includes(id.toLowerCase()); };
var isKnownExceptionIdentifier = function (id) { return LOWERCASE_EXCEPTION_IDS.includes(id.toLowerCase()); };
var parseGplVersionFromIdentifier = function (identifier) {
    var groups = identifier.match(/^([AL]?GPL)\-(\d(\.\d)+)((\-or\-later)|(\-only)|\+)?$/);
    if (groups) {
        return { family: groups[1], version: groups[2], plus: ['-or-later', '+'].includes(groups[4]) };
    }
    return undefined;
};
var expandGplLicenseVersionExpressions = function (relatedLicenseIds) {
    var ids = [];
    relatedLicenseIds.forEach(function (license) {
        ids.push(license);
        if (license.match(/^[AL]?GPL\-\d(\.\d)+$/)) {
            // if "GPL-2.0" or "GPL-3.0", also allow "-only" and "-or-later" variations
            ids.push("".concat(license, "-only"));
            ids.push("".concat(license, "-or-later"));
        }
        var orLaterMatch = license.match(/^([AL]?GPL)\-(\d(\.\d)+)\-or\-later$/);
        if (orLaterMatch) {
            // if "GPL-2.0+" or "GPL-3.0-or-later" etc.
            var licenseFamily_1 = orLaterMatch[1];
            var minimumVersion_1 = orLaterMatch[2];
            var allLicensesInSameFamily = ALL_LICENSE_IDS.filter(function (id) { return id.startsWith("".concat(licenseFamily_1, "-")); });
            var newerLicensesInSameFamily = allLicensesInSameFamily.filter(function (id) {
                var gplId = parseGplVersionFromIdentifier(id);
                return gplId && gplId.version >= minimumVersion_1;
            });
            newerLicensesInSameFamily.forEach(function (id) { return ids.push(id); });
        }
    });
    return ids;
};
var validateLicenseAndExceptionRelation = function (licenseId, exceptionId) {
    var _a;
    var relatedLicenseIds = expandGplLicenseVersionExpressions((((_a = licenses_1.exceptions.exceptions.find(function (x) {
        return x.licenseExceptionId.toLowerCase() === exceptionId.toLowerCase();
    })) === null || _a === void 0 ? void 0 : _a.relatedLicenses) || []));
    if (relatedLicenseIds.length > 0 && !relatedLicenseIds.map(function (x) { return x.toLowerCase(); }).includes(licenseId.toLowerCase())) {
        return ["Exception associated with unrelated license: \"".concat(licenseId, " WITH ").concat(exceptionId, "\" (expected one of: ").concat(relatedLicenseIds.join(', '), ")")];
    }
    return [];
};
var validateLicenseInfoExpression = function (expression) {
    var errors = [];
    var licenseId = expression.license;
    if (!isKnownLicenseIdentifier(licenseId)) {
        errors.push("Unknown SPDX license identifier: ".concat(JSON.stringify(licenseId)));
    }
    if (expression.exception) {
        if (!isKnownExceptionIdentifier(expression.exception)) {
            errors.push("Unknown SPDX exception identifier: ".concat(JSON.stringify(expression.exception)));
        }
        else {
            var relationErrors = validateLicenseAndExceptionRelation(licenseId, expression.exception);
            relationErrors.forEach(function (e) { return errors.push(e); });
        }
    }
    return errors;
};
var validateLicenseRefExpression = function (expression) {
    return []; // LicenseRefs are not currently validated in any way
};
var validateCompoundExpression = function (expression) {
    return validateExpression(expression.left).concat(validateExpression(expression.right));
};
var validateExpression = function (expression) {
    if (expression === null || expression === void 0 ? void 0 : expression.conjunction) {
        return validateCompoundExpression(expression);
    }
    else if (expression === null || expression === void 0 ? void 0 : expression.licenseRef) {
        return validateLicenseRefExpression(expression);
    }
    else if (expression === null || expression === void 0 ? void 0 : expression.license) {
        return validateLicenseInfoExpression(expression);
    }
    /* istanbul ignore next */ (0, assert_1.fail)("Unexpected type of input for validateExpression(".concat(JSON.stringify(expression), ")"));
};
/**
 * Validate the given SPDX expression.
 *
 * @param input The expression {string} to validate.
 * @returns a {ValidationResult} with a binary outcome and a list of validation errors.
 */
function validate(input) {
    if (input.trim() === '') {
        return { valid: false, errors: ['Unknown SPDX identifier: ""'] };
    }
    var parseResult = (0, parser_1.parseSpdxExpressionWithDetails)(input, false);
    if (parseResult.error) {
        return { valid: false, errors: [parseResult.error] };
    }
    else if (parseResult.expression) {
        var errors = validateExpression(parseResult.expression);
        return { errors: errors, valid: errors.length === 0 };
    }
    /* istanbul ignore next */ (0, assert_1.fail)("Unexpected execution path for validate(".concat(JSON.stringify(input), ")"));
}
exports.validate = validate;
exports.default = validate;
